from BaseHTTPServer import BaseHTTPRequestHandler
from GangaCore.Core.GangaRepository import getRegistry, RegistryKeyError
from GangaCore.Core.GangaThread import GangaThread
from GangaCore.Utility.util import hostname
from GangaCore.GPIDev.Base.Proxy import getName
from BaseHTTPServer import HTTPServer

import urlparse
import GangaCore.GPI
from GangaCore.GPIDev.Lib.Config import config
import time
import datetime
import os
logger = GangaCore.Utility.logging.getLogger()

job_status_color = {'new': '00FFFF',
                    'submitting': 'FFFFFF',
                    'submitted': '0000FF',
                    'running': '008000',
                    'completed': '006400',
                    'completing': '006400',
                    'killed': 'FF0000',
                    'unknown': '808080',
                    'incomplete': 'FF00FF',
                    'failed': 'FF0000'}


subjob_status_color = {'new': '00ff7d',
                       'submitting': 'FFFFFF',
                       'submitted': '00007d',
                       'running': '00f000',
                       'completed': '009000',
                       'completing': '009000',
                       'killed': '7d0000',
                       'unknown': '808080',
                       'incomplete': '7d007d',
                       'failed': '7D0000'}


def getColorString(statuses, jobs=True):

    colorString = ""
    colorDictionary = job_status_color

    if not jobs:
        colorDictionary = subjob_status_color

    for status in statuses[:-1]:
        colorString += '%s|' % colorDictionary[status]
    colorString += colorDictionary[statuses[-1]]

    return colorString


def addQuotes(value):

    trimmedValue = value.strip('\'')

    return '"' + trimmedValue + '"'


def get_users_JSON():

    json_users = "{\"basicData\": [[{\"GridName\": \"%s\"}]]}" % config.Configuration.user

    return json_users


def get_subjob_JSON(job):

    result = []
    result.append("{")

    result.append("\"id\": %s," % addQuotes(job.fqid))
    result.append("\"status\": %s," % addQuotes(job.status))
    result.append("\"name\": %s," % addQuotes(job.name))
    result.append("\"application\": %s," %
                  addQuotes(getName(job.application)))
    result.append("\"backend\": %s," %
                  addQuotes(getName(job.backend)))
    result.append("\"actualCE\": %s" % addQuotes(job.backend.actualCE))

    result.append("}")

    return "".join(result)


def get_job_JSON(job):

    undefinedAttribute = 'UNDEFINED'

    result = []
    result.append("{")

    result.append("\"id\": %s," % addQuotes(job.fqid))
    result.append("\"status\": %s," % addQuotes(job.status))
    result.append("\"name\": %s," % addQuotes(job.name))

    # add mon links in the JSON
    mon_links = job.info.monitoring_links
    mon_links_html = ''

    if len(mon_links) > 0:
        number = 1
        for mon_link in mon_links:
            # if it is string -> just the path to the link
            if isinstance(mon_link, str):
                mon_links_html = mon_links_html + \
                    '<div>&nbsp;&nbsp;&nbsp;<a href=\'%s\'>mon_link_%s</a></div>' % (
                        mon_link, number)
                number += 1
            elif isinstance(mon_link, tuple):
                if len(mon_link) == 2:
                    mon_links_html = mon_links_html + \
                        '<div>&nbsp;&nbsp;&nbsp;<a href=\'%s\'>%s</a></div>' % (
                            mon_link[0], mon_link[1])
                else:
                    mon_links_html = mon_links_html + \
                        '<div>&nbsp;&nbsp;&nbsp;<a href=\'%s\'>mon_link_%s</a></div>' % (
                            mon_link[0], number)
                    number += 1

    result.append("\"link\": %s," % addQuotes(mon_links_html))

    # test for expandable data
    result.append("\"inputdir\": %s," % addQuotes(job.inputdir))
    result.append("\"outputdir\": %s," % addQuotes(job.outputdir))

    try:
        result.append("\"submitted\": %s," % addQuotes(
            str(len(job.subjobs.select(status='submitted')))))
        result.append("\"running\": %s," %
                      addQuotes(str(len(job.subjobs.select(status='running')))))
        result.append("\"completed\": %s," % addQuotes(
            str(len(job.subjobs.select(status='completed')))))
        result.append("\"failed\": %s," %
                      addQuotes(str(len(job.subjobs.select(status='failed')))))

        result.append("\"application\": %s," %
                      addQuotes(getName(job.application)))
        result.append("\"backend\": %s," %
                      addQuotes(getName(job.backend)))
        result.append("\"subjobs\": %s," % addQuotes(str(len(job.subjobs))))
        result.append("\"uuid\": %s," % addQuotes(job.info.uuid))

        try:
            result.append("\"actualCE\": %s," %
                          addQuotes(job.backend.actualCE))
        except AttributeError:
            result.append("\"actualCE\": %s," % addQuotes(undefinedAttribute))

    except RegistryKeyError:
        pass

    # remove the last , -> else invalid JSON
    if result[len(result) - 1][-1] == ',':
        last = result[-1]
        result = result[:-1]
        result.append(last[:-1])

    result.append("}")

    return "".join(result)


def get_subjobs_in_time_range(jobid, fromDate=None, toDate=None):
    from GangaCore.Core.GangaRepository import getRegistryProxy

    subjobs = []

    for subjob in getRegistryProxy('jobs')(jobid).subjobs:

        timeCreated = subjob.time.timestamps['new']

        if fromDate is None and toDate is None:

            subjobs.append(subjob)

        elif fromDate is not None and toDate is not None:

            if timeCreated >= fromDate and timeCreated <= toDate:

                subjobs.append(subjob)

        elif fromDate is not None and toDate is None:

            if timeCreated >= fromDate:

                subjobs.append(subjob)

    return subjobs


def get_subjobs_JSON(jobid, fromDate=None, toDate=None):

    json_subjobs_strings = []
    json_subjobs_strings.append("{\"taskjobs\": [")

    subjobs_in_time_range = get_subjobs_in_time_range(jobid, fromDate, toDate)

    for subjob in subjobs_in_time_range:

        json_subjobs_strings.append(get_subjob_JSON(subjob))
        json_subjobs_strings.append(",")

    if json_subjobs_strings[-1] == ",":
        json_subjobs_strings = json_subjobs_strings[:-1]

    json_subjobs_strings.append("]}")

    return "".join(json_subjobs_strings)


def get_job_infos_in_time_range(fromDate=None, toDate=None):

    job_infos = []

    for jobInfo in jobs_dictionary.values():

        timeCreated = jobInfo.getTimeCreated()

        if timeCreated is None:

            if fromDate is None and toDate is None:

                job_infos.append(jobInfo)

        elif fromDate is None and toDate is None:

            job_infos.append(jobInfo)

        elif fromDate is not None and toDate is not None:

            if timeCreated >= fromDate and timeCreated <= toDate:

                job_infos.append(jobInfo)

        elif fromDate is not None and toDate is None:

            if timeCreated >= fromDate:

                job_infos.append(jobInfo)

    return job_infos

# increment dictionary value method


def increment(d, k):
    d.setdefault(k, 0)
    d[k] += 1


def get_accumulated_subjobs_JSON(subjobs):

    completed_dates = []

    for subjob in subjobs:
        if subjob.status == 'completed':
            completed_dates.append(subjob.time.timestamps['final'])

    if len(completed_dates) == 0:
        return ''

    completed_dates.sort()

    start_date = completed_dates[0]
    end_date = completed_dates[-1]

    interval = (end_date - start_date).seconds

    if interval == 0:
        interval = 1

    ratio = 100.0 / interval

    seconds_from_start = []

    scale = len(completed_dates) / 20

    for completed_date in completed_dates:
        seconds_from_start.append(
            ((completed_date - start_date).seconds) * ratio)

    values = []
    for i in range(len(completed_dates)):
        values.append(i + 1)

    reduced_values = []
    reduced_seconds_from_start = []

    ratio1 = 100.0 / len(completed_dates)

    for i in range(len(values)):
        if scale == 0:
            reduced_values.append(str(values[i] * ratio1))
            reduced_seconds_from_start.append(str(seconds_from_start[i]))
        elif (i % scale == 0):
            reduced_values.append(str(values[i] * ratio1))
            reduced_seconds_from_start.append(str(seconds_from_start[i]))

    reduced_values.append(str(values[-1] * ratio1))
    reduced_seconds_from_start.append(str(seconds_from_start[-1]))

    if interval == 1:
        reduced_values = ["0", "100"]
        reduced_seconds_from_start = ["0", "100"]

    returnJSON = "{\"chxl\":\"0:|" + start_date.strftime("%Y-%m-%d %H:%M:%S") + "|" + end_date.strftime("%Y-%m-%d %H:%M:%S") + "\"," + "\"chd\":\"t:" + ','.join(
        reduced_seconds_from_start) + "|" + ','.join(reduced_values) + "\",\"chxr\":\"1,0," + str(values[-1]) + "\"}"

    return returnJSON


def create_subjobs_graphics(jobid, subjob_attribute, fromDate, toDate):

    subjobs_in_time_range = get_subjobs_in_time_range(jobid, fromDate, toDate)

    if subjob_attribute == 'accumulate':
        # return some JSON here
        return get_accumulated_subjobs_JSON(subjobs_in_time_range)

    subjobs_attributes = {}

    for subjob in subjobs_in_time_range:

        if subjob_attribute == 'status':
            increment(subjobs_attributes, subjob.status)

        elif subjob_attribute == 'application':
            increment(
                subjobs_attributes, getName(subjob.application))

        elif subjob_attribute == 'backend':
            increment(subjobs_attributes, getName(subjob.backend))

        elif subjob_attribute == 'actualCE':
            increment(subjobs_attributes, subjob.backend.actualCE)

    if subjob_attribute == 'status':
        return get_pie_chart_json(subjobs_attributes, colors=True, jobs=False)
    else:
        return get_pie_chart_json(subjobs_attributes)


def get_pie_chart_json(d, colors=False, jobs=False):

    #template = "{\"chd\":\"t:50,50\",\"chl\":\"Hello|World\"}"

    if len(d) == 0:
        return "{\"chd\":\"t:1\",\"chl\":\"no data\"}"

    keys = []
    values = []

    for k, v in d.iteritems():
        keys.append(k)
        values.append(v)

    keyString = ""

    for key in keys[:-1]:
        keyString += '%s|' % key
    keyString += keys[-1]

    valueString = ""

    for value in values[:-1]:
        valueString += '%s,' % str(value)
    valueString += str(values[-1])

    result_json = ""

    if colors:
        colorString = getColorString(keys, jobs)
        result_json = "{\"chd\":\"t:%s\",\"chl\":\"%s\",\"chco\":\"%s\"}" % (
            valueString, keyString, colorString)
    else:
        result_json = "{\"chd\":\"t:%s\",\"chl\":\"%s\"}" % (
            valueString, keyString)

    return result_json


def create_jobs_graphics(job_attribute, fromDate=None, toDate=None):

    job_infos_in_time_range = get_job_infos_in_time_range(fromDate, toDate)

    jobs_attribute = {}

    for jobInfo in job_infos_in_time_range:

        if job_attribute == 'status':
            increment(jobs_attribute, jobInfo.getJobStatus())
        elif job_attribute == 'application':
            increment(jobs_attribute, jobInfo.getJobApplication())
        elif job_attribute == 'backend':
            increment(jobs_attribute, jobInfo.getJobBackend())

    if job_attribute == 'status':
        return get_pie_chart_json(jobs_attribute, colors=True, jobs=True)
    else:
        return get_pie_chart_json(jobs_attribute)


def get_jobs_JSON(fromDate=None, toDate=None):

    json_jobs_strings = []
    json_jobs_strings.append("{\"user_taskstable\": [")

    job_infos_in_time_range = get_job_infos_in_time_range(fromDate, toDate)

    for jobInfo in job_infos_in_time_range:

        json_jobs_strings.append(jobInfo.getJobJSON())
        json_jobs_strings.append(",")

    if json_jobs_strings[-1] == ",":
        json_jobs_strings = json_jobs_strings[:-1]

    json_jobs_strings.append("]}")

    return "".join(json_jobs_strings)


def update_jobs_dictionary():
    from GangaCore.Core.GangaRepository import getRegistryProxy

    reg = getRegistry("jobs")
    # get the changed jobs
    changed_ids = reg.pollChangedJobs("WebGUI")

    for job_id in changed_ids:
        try:

            job = getRegistryProxy('jobs')(job_id)

            try:
                jobs_dictionary[job_id] = JobRelatedInfo(
                    job, job.time.timestamps['new'])
            except RegistryKeyError:
                jobs_dictionary[job_id] = JobRelatedInfo(job, None)

        except RegistryKeyError:

            del jobs_dictionary[job_id]


def fill_jobs_dictionary():
    from GangaCore.Core.GangaRepository import getRegistryProxy

    for job in getRegistryProxy('jobs'):
        try:
            # get the id -> it could cause RegistryKeyError and the code below
            # will not be executed
            jobid = job.id

            try:
                jobs_dictionary[jobid] = JobRelatedInfo(
                    job, job.time.timestamps['new'])
            except RegistryKeyError:
                jobs_dictionary[jobid] = JobRelatedInfo(job, None)

        except RegistryKeyError:
            pass

# todo remove


def saveProcessDetails():

    if not os.path.exists(tempFilePath):
        file = open(tempFilePath, 'w')
        try:
            file.write(str(os.getpid()))
            file.write('\n')
            file.write(GangaCore.Utility.util.hostname())
        finally:
            file.close()
# todo remove


def getProcessDetails():

    file = open(tempFilePath, 'r')
    try:
        lines = file.readlines()
        pid = lines[0].strip()
        hostname = lines[1].strip()

        return (pid, hostname)
    finally:
        file.close()


def getHttpServer():

    success = False
    port = httpServerStartTryPort
    server = None

    while not success:

        try:
            server = HTTPServer((httpServerHost, port), GetHandler)
            success = True
        except Exception:
            port += 1

    return server, port


def convertStringToDatetime(timestring):

    time_format = "%Y-%m-%d %H:%M"
    return datetime.datetime.fromtimestamp(time.mktime(time.strptime(timestring, time_format)))


def getFromDateFromTimeRange(timeRange):

    fromDay = None
    today = datetime.date.today()

    if timeRange == 'lastDay':
        fromDay = today
    elif timeRange == 'last2Days':
        fromDay = today + datetime.timedelta(days=-1)
    elif timeRange == 'last3Days':
        fromDay = today + datetime.timedelta(days=-2)
    elif timeRange == 'lastWeek':
        fromDay = today + datetime.timedelta(days=-6)
    elif timeRange == 'last2Weeks':
        fromDay = today + datetime.timedelta(days=-13)
    elif timeRange == 'lastMonth':
        fromDay = today + datetime.timedelta(days=-30)

    fromDateTime = datetime.datetime(
        fromDay.year, fromDay.month, fromDay.day, 0, 0)
    return fromDateTime


def getMonitoringLink(port):

    webMonitoringLink = os.path.join(config['System'][
                                     'GANGA_PYTHONPATH'], 'Ganga', 'Core', 'WebMonitoringGUI', 'client', 'index.html')

    return 'file://' + webMonitoringLink + '?port=' + str(port) + '#user=' + config.Configuration.user + '&timeRange='


class JobRelatedInfo:

    def __init__(self, job, time_created):

        self.job_json = get_job_JSON(job)
        self.time_created = time_created
        self.job_status = job.status
        self.job_application = getName(job.application)
        self.job_backend = getName(job.backend)

    def getJobJSON(self):

        return self.job_json

    def getTimeCreated(self):

        return self.time_created

    def getJobStatus(self):

        return self.job_status

    def getJobApplication(self):

        return self.job_application

    def getJobBackend(self):

        return self.job_backend

    def __hash__(self):

        return hash(self.job_json) + hash(self.time_created)

    def __eq__(self, other):

        return isinstance(other, JobRelatedInfo) and self.job_json == other.job_json and self.time_created == other.time_created


class HTTPServerThread(GangaThread):

    def __init__(self, name):
        GangaThread.__init__(self, name=name)

    def run(self):

        server, port = getHttpServer()
        server.socket.settimeout(1)
        reg = getRegistry("jobs")

        """
        try:
            server = HTTPServer(('pclcg35.cern.ch', 1234), GetHandler)
            server.socket.settimeout(1)
        except Exception:
            return
        
            print("Another Ganga session is already started with --webgui option" )      
            process_details = getProcessDetails()

            print(reg.repository.get_other_sessions())
            print("Process id : %s, hostname : %s" % (process_details[0], process_details[1]))
            
            self.stop()
            self.unregister()   
            return      
        """

        logger.info('Starting web gui monitoring server, please wait ...')

        #   initialization

        reg = getRegistry("jobs")
        # calling here first time will take all jobs
        reg.pollChangedJobs("WebGUI")
        # fill jobs dictionary at the begining
        fill_jobs_dictionary()

        logger.info('Web gui monitoring server started successfully')
        logger.info('You can monitor your jobs at the following location: ' + getMonitoringLink(port))

        # server.serve_forever()

        try:
            while not self.should_stop():
                server.handle_request()
        finally:
            pass
            #print("stopping HTTP server thread")
            # os.remove(tempFilePath)
            # server.server_close()


class GetHandler(BaseHTTPRequestHandler):

    def log_message(self, format, *args):

        logger.debug(format % args)

    def do_GET(self):
        queryString = self.path.split('?')[1]
        import cgi
        qsDict = dict(cgi.parse_qsl(queryString))
        query = qsDict['list']

        fromDate = None
        toDate = None

        # from and to date are either both selected or both not selected
        if 'from' in qsDict and 'to' in qsDict:
            fromDate = convertStringToDatetime(qsDict['from'])
            toDate = convertStringToDatetime(qsDict['to'])
        # if from and to are not selected, it could be timeRange selected
        elif 'timerange' in qsDict:
            fromDate = getFromDateFromTimeRange(qsDict['timerange'])

        json = ''

        if query == "users":
            json = get_users_JSON()
        elif query == "jobs":
            # update dictionary with the changed jobs
            update_jobs_dictionary()
            json = get_jobs_JSON(fromDate, toDate)

        elif query == "subjobs":
            jobid = int(qsDict['taskmonid'])
            json = get_subjobs_JSON(jobid, fromDate, toDate)

        elif query == "jobs_statuses":
            # update dictionary with the changed jobs
            update_jobs_dictionary()
            json = create_jobs_graphics('status', fromDate, toDate)

        elif query == "jobs_backends":
            # update dictionary with the changed jobs
            update_jobs_dictionary()
            json = create_jobs_graphics('backend', fromDate, toDate)

        elif query == "jobs_applications":
            # update dictionary with the changed jobs
            update_jobs_dictionary()
            json = create_jobs_graphics('application', fromDate, toDate)

        elif query == "subjobs_statuses":
            jobid = int(qsDict['taskmonid'])
            json = create_subjobs_graphics(jobid, 'status', fromDate, toDate)

        elif query == "subjobs_backends":
            jobid = int(qsDict['taskmonid'])
            json = create_subjobs_graphics(jobid, 'backend', fromDate, toDate)

        elif query == "subjobs_applications":
            jobid = int(qsDict['taskmonid'])
            json = create_subjobs_graphics(
                jobid, 'application', fromDate, toDate)

        elif query == "subjobs_actualCE":
            jobid = int(qsDict['taskmonid'])
            json = create_subjobs_graphics(jobid, 'actualCE', fromDate, toDate)

        elif query == "subjobs_accumulate":
            jobid = int(qsDict['taskmonid'])
            json = create_subjobs_graphics(
                jobid, 'accumulate', fromDate, toDate)

        elif query == "testaccumulation":

            json = "{\"totaljobs\": [[{\"TOTAL\": 92}], {\"taskmonid\": \"ganga:e60e5904-e63e-432f-b3df-63ca833cf080:\"}], \"procevents\": [[{\"NEventsPerJob\": 0}], {\"taskmonid\": \"ganga:e60e5904-e63e-432f-b3df-63ca833cf080:\"}], \"succjobs\": [[{\"TOTAL\": 92, \"TOTALEVENTS\": 1365491}], {\"taskmonid\": \"ganga:e60e5904-e63e-432f-b3df-63ca833cf080:\"}], \"meta\": {\"genactivity\": null, \"submissiontype\": null, \"site\": null, \"ce\": null, \"dataset\": null, \"submissiontool\": null, \"fail\": null, \"check\": [\"submitted\"], \"date1\": [\"2010-09-23 15:56:27\"], \"date2\": [\"2010-09-24 15:56:27\"], \"application\": null, \"rb\": null, \"status\": null, \"taskmonid\": [\"ganga:e60e5904-e63e-432f-b3df-63ca833cf080:\"], \"args\": \"<![CDATA[taskmonid=ganga%3Ae60e5904-e63e-432f-b3df-63ca833cf080%3A]]>\", \"grid\": null, \"user\": null, \"task\": null, \"unixname\": null, \"sortby\": [\"activity\"], \"activity\": null, \"exitcode\": null}, \"allfinished\": [[{\"finished\": \"2010-08-13 14:02:18\", \"Events\": 2000}, {\"finished\": \"2010-08-13 14:39:13\", \"Events\": 14997}, {\"finished\": \"2010-08-13 14:39:25\", \"Events\": 14350}, {\"finished\": \"2010-08-13 14:39:58\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:40:03\", \"Events\": 14997}, {\"finished\": \"2010-08-13 14:40:18\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:40:19\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:40:37\", \"Events\": 14997}, {\"finished\": \"2010-08-13 14:40:38\", \"Events\": 14994}, {\"finished\": \"2010-08-13 14:40:52\", \"Events\": 14997}, {\"finished\": \"2010-08-13 14:40:53\", \"Events\": 14996}, {\"finished\": \"2010-08-13 14:40:54\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:41:25\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:41:27\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:41:29\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:41:32\", \"Events\": 14997}, {\"finished\": \"2010-08-13 14:41:32\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:41:34\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:41:35\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:41:43\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:41:44\", \"Events\": 14996}, {\"finished\": \"2010-08-13 14:41:45\", \"Events\": 14997}, {\"finished\": \"2010-08-13 14:41:53\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:41:54\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:41:55\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:41:55\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:41:55\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:41:55\", \"Events\": 14997}, {\"finished\": \"2010-08-13 14:41:55\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:41:59\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:42:03\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:42:03\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:42:04\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:42:06\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:42:07\", \"Events\": 14997}, {\"finished\": \"2010-08-13 14:42:14\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:42:14\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:42:27\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:42:27\", \"Events\": 14995}, {\"finished\": \"2010-08-13 14:42:28\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:42:38\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:42:53\", \"Events\": 14997}, {\"finished\": \"2010-08-13 14:42:54\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:42:57\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:42:57\", \"Events\": 14995}, {\"finished\": \"2010-08-13 14:42:58\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:42:58\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:43:01\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:43:02\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:43:04\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:43:04\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:43:11\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:43:15\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:43:15\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:43:17\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:43:22\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:43:23\", \"Events\": 14996}, {\"finished\": \"2010-08-13 14:43:24\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:43:25\", \"Events\": 14996}, {\"finished\": \"2010-08-13 14:43:28\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:43:32\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:43:36\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:43:36\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:43:39\", \"Events\": 14996}, {\"finished\": \"2010-08-13 14:43:43\", \"Events\": 14996}, {\"finished\": \"2010-08-13 14:43:56\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:43:57\", \"Events\": 14299}, {\"finished\": \"2010-08-13 14:43:57\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:44:04\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:44:15\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:44:15\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:44:34\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:44:35\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:44:35\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:44:35\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:44:36\", \"Events\": 14995}, {\"finished\": \"2010-08-13 14:45:03\", \"Events\": 14997}, {\"finished\": \"2010-08-13 14:45:10\", \"Events\": 14998}, {\"finished\": \"2010-08-13 14:45:25\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:45:26\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:45:45\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:45:50\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:45:50\", \"Events\": 14997}, {\"finished\": \"2010-08-13 14:46:01\", \"Events\": 14997}, {\"finished\": \"2010-08-13 14:46:07\", \"Events\": 14996}, {\"finished\": \"2010-08-13 14:46:14\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:46:23\", \"Events\": 14999}, {\"finished\": \"2010-08-13 14:46:26\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:46:30\", \"Events\": 15000}, {\"finished\": \"2010-08-13 14:47:09\", \"Events\": 14999}, {\"finished\": \"2010-08-13 15:57:09\", \"Events\": 14996}, {\"finished\": \"2010-08-13 16:17:45\", \"Events\": 14997}], {\"taskmonid\": \"ganga:e60e5904-e63e-432f-b3df-63ca833cf080:\"}], \"lastfinished\": [[{\"finished\": \"2010-08-13 16:17:45\"}], {\"taskmonid\": \"ganga:e60e5904-e63e-432f-b3df-63ca833cf080:\"}], \"firststarted\": [[{\"started\": \"2010-08-13 13:51:21\"}], {\"taskmonid\": \"ganga:e60e5904-e63e-432f-b3df-63ca833cf080:\"}]}"

        self.send_response(200)
        self.send_header('Content-Type', 'text/html')
        self.end_headers()

        jsonp_function = qsDict['jsonp_callback']
        result = "%s(%s);" % (jsonp_function, json)
        self.wfile.write(result)

        return

jobs_dictionary = {}
httpServerHost = 'localhost'
httpServerStartTryPort = 8080

# todo remove
#import os
#tempFilePath = os.path.join(config.Configuration.gangadir, 'process')
# saveProcessDetails()
# end remove


def start_server():

    t = HTTPServerThread("HTTP_monitoring")
    t.start()

if __name__ == '__main__':
    start_server()



def getEnvironment(c):
    import PACKAGE
    PACKAGE.standardSetup()
    return {}
    
def loadPlugins(c):

    import os,sys
    from GangaCore.Utility.logging import getLogger
    logger = getLogger()
    logger.info('You are now using Python %s',sys.version.split()[0])

    import GangaPlotter.Plotter
    from Plotter.GangaPlotter import GangaPlotter
    #GangaPlotter.Plotter.plotter = GangaPlotter()
    plotter = GangaPlotter()

    from GangaCore.Runtime.GPIexport import exportToGPI

    #exportToGPI('plotter',GangaPlotter.Plotter.plotter,'Objects','graphics plotter')
    exportToGPI('plotter',plotter,'Objects','graphics plotter')



################################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: PrimeTableDataset.py,v 1.1 2008-07-17 16:41:37 moscicki Exp $
################################################################################

from GangaCore.GPIDev.Schema import *
from GangaCore.Utility.Config import getConfig
from GangaCore.GPIDev.Lib.File import File
from GangaCore.GPIDev.Lib.Dataset import Dataset
from GangaCore.Utility.logging import getLogger

class PrimeTableDataset(Dataset):
    """Prime number lookup table definition."""
    
    _schema = Schema(Version(1,0), {
        'table_id_lower': SimpleItem(defvalue=1, doc='The lower bound id of the lookup tables (inclusive).'),
        'table_id_upper': SimpleItem(defvalue=1, doc='The upper bound id of the lookup tables (inclusive).'),
        'table_id_max': SimpleItem(defvalue=15, hidden=1, doc='The upper bound id of the lookup tables (inclusive).'),
        'table_location': SimpleItem(defvalue='http://cern.ch/diane/download/primes', doc='The location of the lookup tables.')
        })
    
    _category = 'datasets'
    _name = 'PrimeTableDataset'
    _exportmethods = ['get_dataset']

    def __init__(self):
        super(PrimeTableDataset,self).__init__()

    def get_dataset(self):
        from urlparse import urlparse
        tables = []

        ## switching lower and upper if different order is given
        if self.table_id_lower > self.table_id_upper:
            tmp = self.table_id_lower
            self.table_id_lower = self.table_id_upper
            self.table_id_upper = tmp

        ## restrict maximum table id to self.table_id_max 
        if self.table_id_upper > self.table_id_max:
            logger.warning('table_id_upper is restricted to up to %d' % self.table_id_max)
            self.table_id_upper = self.table_id_max

        for i in range(self.table_id_lower,self.table_id_upper+1):
            location = '%s/primes%d.zip' % (self.table_location,i)
            url = urlparse(location)
            if url[0] in ['file','']:
                # create new File object if the location is local  
                tables.append(File(location))
            else:    
                tables.append(location)
        return tables

logger = getLogger()



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
"""Write a script containing CMT command which can subsequence be executed."""
import GangaCore.Utility.logging
import os
import sys
import time
import types
import warnings
import tempfile
import shutil
from GangaCore.Utility.Shell import Shell
from AppsBaseUtils import available_packs

logger = GangaCore.Utility.logging.getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def parse_master_package(mstrpckg):
    # first check if we have slashes
    if mstrpckg.find('/') >= 0:
        list = mstrpckg.split('/')
        if len(list) == 3:
            return list
        elif len(list) == 2:
            list.insert(0, '')
            return list
        else:
            raise ValueError("wrongly formatted masterpackage")
    elif mstrpckg.find(' ') >= 0:
        list = mstrpckg.split()
        if len(list) == 3:
            list = (list[2], list[0], list[1])
            return list
        elif len(list) == 2:
            list = ('', list[0], list[1])
            return list
        else:
            raise ValueError("wrongly formatted masterpackage")
    else:
        raise ValueError("wrongly formatted masterpackage")


def CMTscript(app, command=''):
    """Function to execute a cmt command for a specific job. Returns the unix
       exit code.

       Arguments:
          app       - The Gaudi application object to take information from
          command   - String [default ''] The cmt command to execute.
    """
    cmtcmd = 'cmt'
    warnings.filterwarnings('ignore', 'tempnam', RuntimeWarning)

    use = ''
    if app.masterpackage:
        (pack, alg, ver) = parse_master_package(app.masterpackage)
        use = '--use "%s %s %s"' % (alg, ver, pack)

    ura = app.user_release_area
    if not ura:
        expanded = os.path.expandvars("$User_release_area")
        if expanded == "$User_release_area":
            ura = ""
        else:
            ura = expanded.split(os.pathsep)[0]

    cmtoption = ''

    # generate shell script
    script = '#!/bin/sh\n'
    script += 'unalias -a\n'
    script += '. `which LbLogin.sh` -c ' + str(app.platform) + '\n'
    script += 'export User_release_area=' + str(ura) + '\n'
    script += 'unset CMTPROJECTPATH\n'
    script += '. setenvProject.sh '
    setupProjectOptions = ''
    if app.setupProjectOptions:
        setupProjectOptions = app.setupProjectOptions
    script += '%s %s %s %s\n' % (use,
                                 setupProjectOptions, app.appname, app.version)
    command = command.replace('###CMT###', cmtcmd + ' ' + cmtoption)
    logger.debug('Will execute the command: ' + command)
    script += command + '\n'
    logger.debug('The full script for execution:\n' + script)

    # write file
    try:
        tmppath = tempfile.mkdtemp()
        fn = os.path.join(tmppath, 'cmtcommand_script')
        file1 = open(fn, 'w')
    except Exception as e:
        logger.error("Can not create temporary file %s", fn)
        return
    else:
        try:
            file1.write(script)
        finally:
            file1.close()

    # make file executable
    os.chmod(fn, 0o777)
    shell = Shell()
    rc = shell.system(fn)
    if os.path.exists(tmppath):
        shutil.rmtree(tmppath)

    return rc

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



from __future__ import print_function
from __future__ import absolute_import
##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: VPrinterOld.py,v 1.1 2008-07-17 16:40:52 moscicki Exp $
##########################################################################
from GangaCore.GPIDev.Base.Proxy import isProxy, isType, runProxyMethod


def quoteValue(value, selection):
    """A quoting function. Used to get consistent formatting"""
    # print "Quoting",repr(value),selection
    if isType(value, type('')):
        if selection == 'copyable':

            value = value.replace('"', R'\"')
            value = value.replace("'", R"\'")

# DISABLED
##             valueList = list( value )
# for i in range( len( value ) ):
##                 c = value[ i ]
# if c in [ "'", '"' ]:
##                     valueList[ i ] = "\\" + c
##                     value = "".join( valueList )
            if 1 + value.find("\n"):
                # print 'Quote result',"'''" + value + "'''"
                return "'''" + value + "'''"
        # print 'Quote result',"'"+value+"'"
        return "'" + value + "'"
    # print 'Quote result',value
    return value


def indent(level):
    return ' ' * int((level - 1) * 3)


# A visitor to print the object tree.


class VPrinterOld(object):
    # Arguments:
    # out: file-like output stream where to print, default sys.stdout
    # selection: string specifying properties to print (default ''):
    #            'all'            - print all properties
    #            'copyable'       - print only copyable properties
    #            any other string - print unhidden properties

    def __init__(self, out=None, selection=''):
        self.level = 0
        self.nocomma = 1
        self.selection = selection
        if out:
            self.out = out
        else:
            import sys
            self.out = sys.stdout

        # detect whether the body is empty to handle the comma correctly in
        # this case too
        self.empty_body = 0

    def indent(self):
        return indent(self.level)

    def comma(self, force=0):
        if not self.nocomma or force:
            print(",", file=self.out)

        self.nocomma = 0

    def nodeBegin(self, node):
        self.level += 1
        if node._schema is not None:
            print(node._schema.name, '(', file=self.out)
        else:
            print('(', file=self.out)
        self.nocomma = 1
        self.empty_body = 1

    def nodeEnd(self, node):

        if self.empty_body:
            print(self.indent(), ' )', end=' ', file=self.out, sep='')
            self.nocomma = 0
        else:
            if self.nocomma:
                print(')', end=' ', file=self.out)
            else:
                print('\n', self.indent(), ' )', end=' ', file=self.out, sep='')

        self.level -= 1
        if self.level == 0:
            print('\n', file=self.out)

    def showAttribute(self, node, name):
        visible = False
        if self.selection == 'all':
            visible = True
        elif self.selection == 'copyable':
            if node._schema.getItem(name)['copyable']:
                if not node._schema.getItem(name)['hidden']:
                    visible = True
        elif self.selection == 'preparable':
            # the following relies on the assumption that we only ever call printPrepTree on
            # a preparable application.
            if node._schema.getItem(name)['preparable'] or self.level == 2:
                if not node._schema.getItem(name)['hidden']:
                    visible = True
        else:
            if not node._schema.getItem(name)['hidden']:
                visible = True
        return visible

    def simpleAttribute(self, node, name, value, sequence):
        if self.showAttribute(node, name):
            self.empty_body = 0
            self.comma()
            # DISABLED
            # print '*'*20,name
            # if sequence:
            #    print 'transformation:',repr(value)
            #    value = value.toString()
            #    print 'into',repr(value)
            # else:
            #    print 'no transformation'
            print(self.indent(), name, '=', self.quote(value), end=' ', file=self.out)

    def sharedAttribute(self, node, name, value, sequence):
        self.simpleAttribute(node, name, value, sequence)

    def acceptOptional(self, s):
        if s is None:
            print(None, end=' ', file=self.out)
        else:
            runProxyMethod(s, 'accept', self)

    def componentAttribute(self, node, name, subnode, sequence):
        if self.showAttribute(node, name):
            self.empty_body = 0
            self.comma()
            print(self.indent(), name, '=', end=' ', file=self.out)
            if sequence:
                print('[', end=' ', file=self.out)
                for s in subnode:
                    self.acceptOptional(s)
                    print(',', end=' ', file=self.out)
                print(']', end=' ', file=self.out)
            else:
                self.acceptOptional(subnode)

    def quote(self, x):
        return quoteValue(x, self.selection)





import pickle
import os
import copy
from GangaCore.GPIDev.Adapters.ISplitter import ISplitter
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.files import expandfilename
from GangaCore.GPIDev.Lib.Job import Job
from GangaCore.Utility.logging import getLogger
from GangaLHCb.Lib.Applications.GaudiExec import GaudiExec

logger = getLogger()

class GaussSplitter(ISplitter):

    """Create a set of Gauss jobs based on the total number of jobs and the
    number of events per subjob.

    This Splitter will create a set of Gauss jobs using two parameters:
    'eventsPerJob' and 'numberOfJobs'. Each job uses a different random seed
    using the Gaudi options file statement 'GaussGen.FirstEventNumber' and will
    produce the amount of events sepcified in 'eventsPerJob'. The total number
    of generated events therefore will be 'eventsPerJob*numberOfJob'.
    """
    _name = "GaussSplitter"
    _schema = Schema(Version(1, 0), {
        'eventsPerJob': SimpleItem(defvalue=5, doc='Number of generated events per job'),
        'numberOfJobs': SimpleItem(defvalue=2, doc="No. of jobs to create"),
        'firstEventNumber': SimpleItem(defvalue=0, doc="First event number for first subjob")
    })

    def _create_subjob(self, job, inputdata):
        j = Job()
        j.copyFrom(job)
        j.splitter = None
        j.merger = None
        j.inputsandbox = []  # master added automatically
        j.inputfiles = []
        j.inputdata = inputdata

        return j

    def split(self, job):
        """
            Method to do the splitting work
            Args:
                job (Job): master job to be used as a template to split subjobs
        """

        from GangaLHCb.Lib.Applications import Gauss
        if not isinstance(job.application, (Gauss, GaudiExec)):
            logger.warning("This application is of type: '%s', be careful how you use it with the GaussSplitter!" % type(job.application))

        subjobs = []

        inputdata = job.inputdata

        if not isinstance(job.application, GaudiExec):
            # I'm assuming this file is created by the Gauss Application at some stage?
            if not job.inputdata:
                share_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),
                                          'shared',
                                          getConfig('Configuration')['user'],
                                          job.application.is_prepared.name,
                                          'inputdata',
                                          'options_data.pkl')

                if os.path.exists(share_path):
                    f = open(share_path, 'r+b')
                    #FIXME should this have been an addition?
                    inputdata = pickle.load(f)
                    f.close()

        for i in range(self.numberOfJobs):
            j = self._create_subjob(job, inputdata)
            # FIXME this starts from the 1st event and not zero, is it clear why?
            first = self.firstEventNumber + i * self.eventsPerJob + 1
            opts = 'from Configurables import GenInit \n'
            if isinstance(job.application, Gauss):
                opts += 'from Gaudi.Configuration import * \n'
                opts += 'ApplicationMgr().EvtMax = %d\n' % self.eventsPerJob
            else:
                opts += 'from Configurables import LHCbApp \n'
                opts += 'LHCbApp().EvtMax = %d\n' % self.eventsPerJob
            opts += 'GenInit("GaussGen").FirstEventNumber = %d\n' % first
            spillOver = ["GaussGenPrev", "GaussGenPrevPrev", "GaussGenNext"]
            for s in spillOver:
                opts += 'GenInit("%s").FirstEventNumber = %d\n' % (s, first)
            #j.application.extra.input_buffers['data.py'] += opts
            if isinstance(job.application, GaudiExec):
                j.application.extraOpts = j.application.extraOpts + '\n' + opts
            else:
                j._splitter_data = opts
            # j.inputsandbox.append(File(FileBuffer(path,opts).create().name))
            logger.debug("Creating job %d w/ FirstEventNumber = %d" % (i, first))
            subjobs.append(j)

        return subjobs





""" Refer to Ganga/PACKAGE.py for details on the purpose of this module.
"""

external_packages = {
        #TestOOB testing tool
        'PYTF' : {'version' : '1.6', 'noarch':True, 'PYTHONPATH':'.', 'PYTF_TOP_DIR':'.'},
        #coverage reporting tool
        'figleaf' : {'version' : '0.6', 'noarch':True, 'PYTHONPATH':'.'},
        #test externals
        'test-externals':{'version':'1.0', 'noarch':True, 'TEST_EXTERNAL_PATH':'.'},
#        #subprocess
#        'subprocess':{'version':'2.4.2', 'noarch':True, 'PYTHONPATH':'lib/python2.2/site-packages', 'maxHexVersion':'0x20400f0'}
        }

from GangaCore.Utility.Setup import PackageSetup

setup = PackageSetup(external_packages)

def standardSetup(setup=setup):
        for p in setup.packages:
                setup.prependPath(p,'PYTHONPATH')
                setup.prependPath(p,'LD_LIBRARY_PATH')
                setup.prependPath(p,'PATH')
                setup.setPath(p,'PYTF_TOP_DIR')
                setup.setPath(p,'TEST_EXTERNAL_PATH')




from __future__ import absolute_import
##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: MassStorageFile.py,v 0.1 2011-11-09 15:40:00 idzhunov Exp $
##########################################################################
import errno
import re
import os
import os.path
import copy
import glob
import inspect
import time
from fnmatch import fnmatch
from pipes import quote

from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem

from GangaCore.Utility.Config import getConfig
from GangaCore.GPIDev.Base.Proxy import stripProxy
from GangaCore.Utility import Shell
from GangaCore.Utility.logging import getLogger
from GangaCore.GPIDev.Adapters.IGangaFile import IGangaFile
from GangaCore.GPIDev.Base.Objects import _getName
from GangaCore.Utility.files import expandfilename
from GangaCore.Core.exceptions import GangaException

import GangaCore.Utility.Config

regex = re.compile('[*?\[\]]')
logger = getLogger()

class MassStorageFile(IGangaFile):
    """MassStorageFile represents a class marking a file to be written into mass storage (like Castor at CERN)
    """
    _schema = Schema(Version(1, 1), {'namePattern': SimpleItem(defvalue="", doc='pattern of the file name'),
                                     'localDir': SimpleItem(defvalue="", copyable=1, doc='local dir where the file is stored, used from get and put methods'),
                                     'joboutputdir': SimpleItem(defvalue="", doc='outputdir of the job with which the outputsandbox file object is associated'),
                                     'locations': SimpleItem(defvalue=[], copyable=1, typelist=[str], sequence=1, doc="list of locations where the outputfiles are uploaded"),
                                     'outputfilenameformat': SimpleItem(defvalue=None, typelist=[str, None], protected=0,\
                                                    doc="keyword path to where the output should be uploaded, i.e. /some/path/here/{jid}/{sjid}/{fname},\
                                                        if this field is not set, the output will go in {jid}/{sjid}/{fname} or in {jid}/{fname}\
                                                        depending on whether the job is split or not"),
                                     'inputremotedirectory': SimpleItem(defvalue=None, typelist=[str, None], protected=0, doc="Directory on mass storage where the file is stored"),
                                     'subfiles': ComponentItem(category='gangafiles', defvalue=[], hidden=1, sequence=1, copyable=0,\
                                                    doc="collected files from the wildcard namePattern"),
                                     'failureReason': SimpleItem(defvalue="", protected=1, copyable=0, doc='reason for the upload failure'),
                                     'compressed': SimpleItem(defvalue=False, typelist=[bool], protected=0, doc='wheather the output file should be compressed before sending somewhere')
                                     })

    _category = 'gangafiles'
    _name = "MassStorageFile"
    _exportmethods = ["location", "get", "put", "setLocation", "remove", "accessURL"]

    _additional_slots = ['shell']

    def __init__(self, namePattern='', localDir='', **kwds):
        """
        MassStorageFile construction
        Args:
            namePattern (str): is the pattern of the output file that has to be written into mass storage
            localDir (str): This is the optional local directory of a file to be uploaded to mass storage
        """
        self._checkConfig()
        super(MassStorageFile, self).__init__()
        self._setNamePath(_namePattern=namePattern, _localDir=localDir)
        self.locations = []
        self.shell = Shell.Shell()

    def __setattr__(self, attr, value):
        """
        This is an overloaded setter method to make sure that we're auto-expanding the filenames of files which exist.
        In the case we're assigning any other attributes the value is simply passed through
        Args:
            attr (str): This is the name of the attribute which we're assigning
            value (unknown): This is the value being assigned.
        """
        actual_value = value
        if attr == "namePattern":
            this_localDir, actual_value = os.path.split(value)
            if this_localDir:
                self.localDir = this_localDir
        if attr == "localDir":
            if value and (value.find(':') == -1):
                actual_value = os.path.abspath(expandfilename(value))

        super(MassStorageFile, self).__setattr__(attr, actual_value)

    def _setNamePath(self, _namePattern='', _localDir=''):
        if _namePattern != '' and _localDir == '':
            self.namePattern = os.path.basename(_namePattern)
            if not os.path.dirname(_namePattern):
                if os.path.isfile(os.path.join(os.getcwd(), os.path.basename(_namePattern))):
                    self.localDir = os.getcwd()
            else:
                self.localDir = os.path.dirname(_namePattern)
        elif _namePattern != '' and _localDir != '':
            self.namePattern = _namePattern
            self.localDir = _localDir

    def _checkConfig(self):
        """
        Check that the MassStorageFile configuration is correct
        """
        if not getConfig('Output')[_getName(self)]['uploadOptions']['path'] :
            raise GangaException('Unable to create MassStorageFile. Check your configuration!')

    def __repr__(self):
        """Get the representation of the file."""

        return "%s(namePattern='%s')" % (_getName(self), self.namePattern)

    def mass_line_processor(self, line):
        """ This function splits the input line from the post-processsing system to define where this file is:
        Args:
            line(str): This is expected to be in the format of the postprocessor file from jobs transfering files on the WN
        """
        lineParts = line.split()
        pattern = lineParts[1]
        outputPath = lineParts[2]
        split_name = os.path.splitext(outputPath)
        if split_name[1] == '.gz':
            name = split_name[0]
        else:
            name = outputPath

        if regex.search(self.namePattern) is not None:
            if outputPath == 'ERROR':
                logger.error("Failed to upload file to mass storage")
                logger.error(line[line.find('ERROR') + 5:])
                d = copy.deepcopy(self)
                d.namePattern = pattern
                d.compressed = self.compressed
                d.failureReason = line[line.find('ERROR') + 5:]
                self.subfiles.append(d)
            else:
                if pattern == self.namePattern:
                    d = copy.deepcopy(self)
                    d.namePattern = name
                    self.subfiles.append(d)
                    d.mass_line_processor(line)
        elif name == self.namePattern:
            if outputPath == 'ERROR':
                logger.error("Failed to upload file to mass storage")
                logger.error(line[line.find('ERROR') + 5:])
                self.failureReason = line[line.find('ERROR') + 5:]
                return
        self.locations = [outputPath.strip('\n')]

    def setLocation(self):
        """
        Sets the location of output files that were uploaded to mass storage from the WN
        """
        job = self.getJobObject()

        postprocessLocationsPath = os.path.join(job.outputdir, getConfig('Output')['PostProcessLocationsFileName'])
        if not os.path.exists(postprocessLocationsPath):
            return

        for line in open(postprocessLocationsPath, 'r'):

            if line.strip() == '':
                continue

            if line.startswith('massstorage'):
                self.mass_line_processor(line.strip())

    def location(self):
        """
        Return list with the locations of the post processed files (if they were configured to upload the output somewhere)
        """
        tmpLocations = []
        if self.subfiles:
            for i in self.subfiles:
                tmpLocations.append(i.locations)
        else:
            tmpLocations = self.locations
        return tmpLocations

    def internalCopyTo(self, targetPath):
        """
        Copy a the file to the local storage using the get mechanism
        Args:
            targetPath (str): Target path where the file is to copied to
        """
        to_location = targetPath

        cp_cmd = getConfig('Output')[_getName(self)]['uploadOptions']['cp_cmd']

        for location in self.locations:
            targetLocation = os.path.join(to_location, os.path.basename(location))
            self.execSyscmdSubprocess('%s %s %s' % (cp_cmd, quote(location), quote(targetLocation)))

    def getWNScriptDownloadCommand(self, indent):
        ## FIXME fix me for the situation of multiple files?

        script = """\n

###INDENT###os.system(\'###CP_COMMAND###\')

"""
        cp_cmd = '%s %s .' % (getConfig('Output')[_getName(self)]['uploadOptions']['cp_cmd'], quote(self.locations[0]))

        replace_dict = { '###INDENT###' : indent, '###CP_COMMAND###' : cp_cmd }

        for k, v in replace_dict.iteritems():
            script = script.replace(str(k), str(v))

        return script

    def _mkdir(self, massStoragePath, exitIfNotExist=False):
        """
        Creates a folder on the mass Storage corresponding to the given path
        Args:
            massStoragePath (str): This is the path we want to make if it doesn't exist.
        """

        massStorageConfig = getConfig('Output')[_getName(self)]['uploadOptions']
        mkdir_cmd = massStorageConfig['mkdir_cmd']
        ls_cmd = massStorageConfig['ls_cmd']

        # create the last directory (if not exist) from the config path
        pathToDirName = os.path.dirname(massStoragePath)
        dirName = os.path.basename(massStoragePath)

        directoryExists = False

        (exitcode, mystdout, mystderr) = self.execSyscmdSubprocess('%s %s' % (ls_cmd, quote(pathToDirName)))
        if exitcode != 0 and exitIfNotExist:
            self.handleUploadFailure(mystderr, '1) %s %s' % (ls_cmd, pathToDirName))
            raise GangaException(mystderr)

        for directory in mystdout.split('\n'):
            if directory.strip() == dirName:
                directoryExists = True
                break

        if not directoryExists:
            (exitcode, mystdout, mystderr) = self.execSyscmdSubprocess('%s -p %s' % (mkdir_cmd, quote(massStoragePath)))
            if exitcode != 0:
                self.handleUploadFailure(mystderr, '2) %s %s' % (mkdir_cmd, massStoragePath))
                raise GangaException(mystderr)

    def put(self):
        """
        Creates and executes commands for file upload to mass storage (Castor), this method will
        be called on the client
        """

        sourceDir = ''

        # if used as a stand alone object
        if self._getParent() is None:
            if self.localDir == '':
                _CWD = os.getcwd()
                if os.path.isfile(os.path.join(_CWD, self.namePattern)):
                    sourceDir = _CWD
                else:
                    logger.warning('localDir attribute is empty, don\'t know from which dir to take the file')
                    return
            else:
                sourceDir = self.localDir

                (result, message) = self.validate()

                if result == False:
                    logger.warning(message)
                    return

        else:
            job = self.getJobObject()
            sourceDir = job.outputdir

            # if there are subjobs, the put method will be called on every subjob
            # and will upload the resulted output file
            if len(job.subjobs) > 0:
                return

        massStorageConfig = getConfig('Output')[_getName(self)]['uploadOptions']

        cp_cmd = massStorageConfig['cp_cmd']
        ls_cmd = massStorageConfig['ls_cmd']
        massStoragePath = massStorageConfig['path']

        try:
            self._mkdir(massStoragePath, exitIfNotExist=True)
        except GangaException:
            return

        # the folder part of self.outputfilenameformat
        folderStructure = ''
        # the file name part of self.outputfilenameformat
        filenameStructure = ''

        if not self.outputfilenameformat:
            filenameStructure = '{fname}'

            parent = self._getParent()
            if parent is not None:
                folderStructure = '{jid}'
                if parent._getParent() is not None:
                    folderStructure = os.path.join(folderStructure, '{sjid}')

        else:
            folderStructure = os.path.dirname(self.outputfilenameformat)
            filenameStructure = os.path.basename(self.outputfilenameformat)

        folderStructure = self.expandString(folderStructure)

        # create the folder structure
        if folderStructure:
            massStoragePath = os.path.join(massStoragePath, folderStructure)
            try:
                self._mkdir(massStoragePath)
            except GangaException:
                return

        # here filenameStructure has replaced jid and sjid if any, and only not
        # replaced keyword is fname
        fileName = self.namePattern
        if self.compressed:
            fileName = '%s.gz' % self.namePattern

        if regex.search(fileName) is not None:
            for currentFile in glob.glob(os.path.join(sourceDir, fileName)):
                finalFilename = self.expandString(filenameStructure, os.path.basename(currentFile))
           
                (exitcode, mystdout, mystderr) = self.execSyscmdSubprocess('%s %s %s' %\
                                                (cp_cmd, quote(currentFile), quote(os.path.join(massStoragePath, finalFilename))))

                d = copy.deepcopy(self)
                d.namePattern = os.path.basename(currentFile)
                d.localDir = os.path.dirname(currentFile)
                d.compressed = self.compressed

                if exitcode != 0:
                    self.handleUploadFailure(mystderr, '4) %s %s %s' % (cp_cmd, currentFile, os.path.join(massStoragePath, finalFilename)))
                else:
                    logger.info('%s successfully uploaded to mass storage as %s' % (currentFile, os.path.join(massStoragePath, finalFilename)))
                    d.locations = os.path.join(massStoragePath, os.path.basename(finalFilename))

                self.subfiles.append(d)
        else:
            currentFile = os.path.join(sourceDir, fileName)
            finalFilename = self.expandString(filenameStructure, fileName)
            (exitcode, mystdout, mystderr) = self.execSyscmdSubprocess('%s %s %s' %\
                                                        (cp_cmd, quote(currentFile), quote(os.path.join(massStoragePath, finalFilename))))
            if exitcode != 0:
                self.handleUploadFailure(mystderr, '5) %s %s %s' % (cp_cmd, currentFile, os.path.join(massStoragePath, finalFilename)))
            else:
                logger.info('%s successfully uploaded to mass storage as %s' % (currentFile, os.path.join(massStoragePath, finalFilename)))
                location = os.path.join(massStoragePath, os.path.basename(finalFilename))
                if location not in self.locations:
                    self.locations.append(location)


    def validate(self):

        # if the user has set outputfilenameformat, validate for presence of
        # jid, sjid and fname keywords depending on job type - split or
        # non-split
        if self.outputfilenameformat is not None:

            searchFor = ['{fname}']
            isJob = False
            isSplitJob = False

            if self._getParent() is not None:

                isJob = True

                if stripProxy(self.getJobObject()).master is not None:

                    isSplitJob = True
                    searchFor.append('{sjid}')

            missingKeywords = []

            for item in searchFor:
                if self.outputfilenameformat.find(item) == -1:
                    missingKeywords.append(item)

            if len(missingKeywords):
                return (False, 'Error in %s.outputfilenameformat field : missing keywords %s ' % (_getName(self), ','.join(missingKeywords)))

            if isSplitJob == False and self.outputfilenameformat.find('{sjid}') > -1:
                return (False, 'Error in %s.outputfilenameformat field :  job is non-split, but {\'sjid\'} keyword found' % _getName(self))

            if isJob == False and self.outputfilenameformat.find('{sjid}') > -1:
                return (False, 'Error in %s.outputfilenameformat field :  no parent job, but {\'sjid\'} keyword found' % _getName(self))

            if isJob == False and self.outputfilenameformat.find('{jid}') > -1:
                return (False, 'Error in %s.outputfilenameformat field :  no parent job, but {\'jid\'} keyword found' % _getName(self))

            invalidUnixChars = ['"', ' ']
            test = self.outputfilenameformat.replace('{jid}', 'a').replace('{sjid}', 'b').replace('{fname}', 'c')

            for invalidUnixChar in invalidUnixChars:
                if test.find(invalidUnixChar) > -1:
                    return (False, 'Error in %s.outputfilenameformat field :  invalid char %s found' % (_getName(self), invalidUnixChar))

        return (True, '')

    def handleUploadFailure(self, error, cmd_run_str=''):
        """
        Function to display what went wrong with an associated Job id if there is one and to assign failureReason for future.
        Args:
            error (str): This is the error which was given from the shell command
            cmd_run_str (str): This is a string related to but not always exactly the command run.
        """

        self.failureReason = error
        if self._getParent() is not None:
            logger.error("Job %s failed. One of the job.outputfiles couldn't be uploaded because of %s" % (str(self._getParent().fqid), self.failureReason))
        else:
            logger.error("The file can't be uploaded because of %s" % (self.failureReason))
        if cmd_run_str:
            logger.error("Attempted to run: '%s'" % (cmd_run_str))

    def getWNInjectedScript(self, outputFiles, indent, patternsToZip, postProcessLocationsFP):
        """
        Returns script that have to be injected in the jobscript for postprocessing on the WN
        """
        massStorageCommands = []

        massStorageConfig = getConfig('Output')[_getName(self)]['uploadOptions']

        for outputFile in outputFiles:

            outputfilenameformat = 'None'
            if outputFile.outputfilenameformat is not None and outputFile.outputfilenameformat != '':
                outputfilenameformat = outputFile.outputfilenameformat

            massStorageCommands.append(['massstorage', outputFile.namePattern, outputfilenameformat,
                                        massStorageConfig['mkdir_cmd'],  massStorageConfig['cp_cmd'],
                                        massStorageConfig['ls_cmd'], massStorageConfig['path']])

        script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))),
                                        'scripts/MassStorageFileWNScript.py.template')

        from GangaCore.GPIDev.Lib.File import FileUtils
        script = FileUtils.loadScript(script_location, '###INDENT###')

        jobfqid = self.getJobObject().fqid

        jobid = jobfqid
        subjobid = ''

        if (jobfqid.find('.') > -1):
            jobid = jobfqid.split('.')[0]
            subjobid = jobfqid.split('.')[1]

        replace_dict = {'###MASSSTORAGECOMMANDS###' : repr(massStorageCommands),
                        '###PATTERNSTOZIP###' : str(patternsToZip),
                        '###INDENT###' : indent,
                        '###POSTPROCESSLOCATIONSFP###' : postProcessLocationsFP,
                        '###FULLJOBDIR###' : str(jobfqid.replace('.', os.path.sep)),
                        '###JOBDIR###' : str(jobid),
                        '###SUBJOBDIR###' : str(subjobid)}

        for k, v in replace_dict.iteritems():
            script = script.replace(str(k), str(v))

        return script

    def processWildcardMatches(self):
        if self.subfiles:
            return self.subfiles

        if regex.search(self.namePattern):
            ls_cmd = getConfig('Output')[_getName(self)]['uploadOptions']['ls_cmd']
            exitcode, output, m = self.shell.cmd1(ls_cmd + ' ' + self.inputremotedirectory, capture_stderr=True)

            for filename in output.split('\n'):
                if fnmatch(filename, self.namePattern):
                    subfile = copy.deepcopy(self)
                    subfile.namepattern = filename
                    subfile.inputremotedirectory = self.inputremotedirectory

                    self.subfiles.append(subfile)

    def remove(self, force=False, removeLocal=False):
        """
        Removes file from remote storage ONLY by default
        """
        massStorageConfig = getConfig('Output')[_getName(self)]['uploadOptions']
        rm_cmd = massStorageConfig['rm_cmd']

        if force == True:
            _auto_delete = True
        else:
            _auto_delete = False

        for i in self.locations:

            if not _auto_delete:

                keyin = None

                while keyin is None:
                    keyin = raw_input("Do you want to delete file %s at Location: %s ? [y/n] " % (str(self.namePattern), str(i)))
                    if keyin.lower() == 'y':
                        _delete_this = True
                    elif keyin.lower() == 'n':
                        _delete_this = False
                    else:
                        logger.warning("y/n please!")
                        keyin = None
            else:
                _delete_this = True

            if _delete_this:
                logger.info("Deleting File at Location: %s")
                self.execSyscmdSubprocess('%s %s' % (rm_cmd, quote(i)))
                self.locations.pop(i)

        if removeLocal:

            sourceDir = ''
            if self.localDir == '':
                _CWD = os.getcwd()
                if os.path.isfile(os.path.join(_CWD, self.namePattern)):
                    sourceDir = _CWD
            else:
                sourceDir = self.localDir

            _localFile = os.path.join(sourceDir, self.namePattern)

            if os.path.isfile(_localFile):

                if force:
                    _actual_delete = True
                else:

                    keyin = None
                    while keyin is None:
                        keyin = raw_input("Do you want to remove the local File: %s ? ([y]/n) " % str(_localFile))
                        if keyin.lower() in ['y', '']:
                            _actual_delete = True
                        elif keyin.lower() == 'n':
                            _actual_delete = False
                        else:
                            logger.warning("y/n please!")
                            keyin = None

                if _actual_delete:
                    remove_filename = _localFile + "_" + str(time.time()) + '__to_be_deleted_'

                    try:
                        os.rename(_localFile, remove_filename)
                    except OSError as err:
                        logger.warning("Error in first stage of removing file: %s" % remove_filename)
                        remove_filename = _localFile

                    try:
                        os.remove(remove_filename)
                    except OSError as err:
                        if err.errno != errno.ENOENT:
                            logger.error("Error in removing file: %s" % str(remove_filename))
                            raise
                        pass
        return

    def accessURL(self):

        # Need to come up with a prescription based upon the server address and
        # file on EOS or elsewhere to return a full URL which we can pass to
        # ROOT...

        protoPath = getConfig('Output')[_getName(self)]['defaultProtocol']

        myLocations = self.location()

        accessURLs = []

        for _file in myLocations:
            accessURLs.append(protoPath + os.path.join(os.sep, _file))

        return accessURLs

# add MassStorageFile objects to the configuration scope (i.e. it will be
# possible to write instatiate MassStorageFile() objects via config file)
GangaCore.Utility.Config.config_scope['MassStorageFile'] = MassStorageFile

class SharedFile(MassStorageFile):

    ''' SharedFile. Special case of MassStorage for locally accessible fs through the standard lsb commands. '''
    _schema = MassStorageFile._schema.inherit_copy()
    _category = 'gangafiles'
    _name = 'SharedFile'

    # Copied from MassStorageFile to keep interface
    def __init__(self, namePattern='', localDir='', **kwds):
        """
        SharedFile construction
        Args:
            namePattern (str): is the pattern of the output file that has to be written into mass storage
            localDir (str): This is the optional local directory of a file to be uploaded to mass storage
        """
        if getConfig('Output')[_getName(self)]['uploadOptions']['path'] is None:
            logger.error("In order to use the SharedFile class you will need to define the path directory in your .gangarc")
            raise GangaException("In order to use the SharedFile class you will need to define the path directory in your .gangarc")

        super(SharedFile, self).__init__(namePattern, localDir, **kwds)

# add SharedFile objects to the configuration scope (i.e. it will be
# possible to write instatiate SharedFile() objects via config file)
GangaCore.Utility.Config.config_scope['SharedFile'] = SharedFile




def static_vars(**kwargs):
    """
    Decorate a function and provide it with some static variables.
    >>> @static_vars(counter=0)
    ... def foo():
    ...     foo.counter += 1
    ...     return foo.counter
    >>> foo()
    1
    >>> foo()
    2
    """
    def decorate(func):
        func.__dict__.update(kwargs)
        return func
    return decorate


"""Library of GangaRobot.Framework.Action.IAction implementations.

See Base for abstract submit / finish / extract / report implementations.
See Core for generic submit / finish / extract / report implementations.
See Ext for specialised implementations.
See Example for example implementations.

"""


from os import rename, path, makedirs, chdir, unlink, listdir, chmod
from os import stat as os_stat
import random
import tempfile
import time
import subprocess
import shutil
import tarfile
import threading
import stat
import uuid
from functools import wraps
from StringIO import StringIO

from GangaCore.Core.exceptions import ApplicationConfigurationError, ApplicationPrepareError, GangaException
from GangaCore.GPIDev.Adapters.IGangaFile import IGangaFile
from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Base.Filters import allComponentFilters
from GangaCore.GPIDev.Base.Proxy import getName
from GangaCore.GPIDev.Lib.File.File import ShareDir
from GangaCore.GPIDev.Lib.File.LocalFile import LocalFile
from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem, GangaFileItem
from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.files import expandfilename, fullpath

from GangaDirac.Lib.Files.DiracFile import DiracFile
from GangaDirac.Lib.Backends.DiracBase import DiracBase

from .GaudiExecUtils import getGaudiExecInputData, _exec_cmd, getTimestampContent, gaudiPythonWrapper

logger = getLogger()

def gaudiExecBuildLock(f):
    """ Method used to lock the build methods in GaudiExec so we don't run multiple builds in parallel.
    This is because each new build destorys the target.
    Args:
        f(function): This should be the buildGangaTarget from GaudiExec
    """
    @wraps(f)
    def masterPrepLock(self, *args, **kwds):
        
        # Get the global lock and prepare
        with gaudiExecBuildLock.globalBuildLock:
            return f(self,*args, **kwds)

    return masterPrepLock

# Global lock for all builds
gaudiExecBuildLock.globalBuildLock = threading.Lock()

class GaudiExec(IPrepareApp):
    """

    Welcome to the new GaudiApp for LHCb apps written/constructed making use of the new CMake framework

    =============
    Simple Usage:
    =============
    The simplest usage of GaudiExec can be achieved by using the 'prepareGaudiExec' function

    e.g.

    j=Job(application=prepareGaudiExec('DaVinci','v41r3'));

    This creates a new application env within 'cmtuser' (this is configurable) and returns a GaudiExec object for Ganga to use.
    This is equivalent to running over a released application when you don't want to check out any private code or make any code changes.

    =============
    Requirements:
    =============

    Before submitting jobs with this application you will need to run something similar to the following:
    (outside of Ganga at the command line)

    cd $SOMEPATH
    lb-dev DaVinci v40r2
    cd $SOMEPATH/DaVinciDev_v40r2
    getpack 

    This program will perform the following command to `prepare` the application before submission:
        make ganga-input-sandbox
    NB:
    The output from this command can be quite large and Ganga will save it to disk and store it at least once per (master) job
    If your build target is large I would advise that you consider placing your gangadir in your AFS workspace where there is more storage available

    ======
    Usage:
    ======

    This application needs to be configured with the absolute directory of the project and the options you want to pass to gaudirun.py

    e.g.

    j=Job()
    myApp = GaudiExec()
    myApp.directory = "$SOMEPATH/DaVinciDev_v40r2"
    myApp.options = ["$SOMEPATH/DaVinciDev_v40r2/myDaVinciOpts.py"]
    j.application = myApp
    j.submit()

    To setup a minimal application you can also run the helper function:

    prepare_cmake_app(myApp, myVer, myPath, myGetpack)

    =============
    How it works:
    =============

    The actual command run on the WN is::

        ./run gaudirun.py optionsFile.py data.py

    If you would prefer to have your optsfile run as a python application aka like 'GaudiPython' style jobs.
    Set:
        job.application.useGaudiRun = False

    This then changes the command run on the WN to be::

        ./run python OptsFileWrapper.py

    Here the OptsFileWrapper script imports the extraOpts and the data.py describing the data to be run over and executes options in the global namespace with 'execfile'
    The OptsFileWrapper will _execute_ the first file in the job.application.options and will import all other opts files before executing this one.

    =================
    Get job metadata:
    =================

    To get the Gaudi metadata from the job set the option 'getMatadata' to True - the default is false.

    j.application.getMetadata = True

    This will add some options to running the job to create a summary.xml file which is downloaded in the output sandbox and parsed by ganga. Ganga will also
    merge the summary.xml files using the Gaudi XML merging script for each subjob to create the metadata for the whole job.

    ========================
    Use subjob id in options
    ========================

    The GaudiExec application sets an environment variable 'ganga_jobid' when it runs. This is the id of the job. This can then be used in your job options:

    import os
    jobid = os.getenv('ganga_jobid')

    For example the outputfile of each subjob can be labelled with the subjob number, useful in case you want to further process your outputfiles on the grid.

    """
    _schema = Schema(Version(1, 0), {
        # Options created for constructing/submitting this app
        'directory':    SimpleItem(defvalue='', typelist=[None, str], comparable=1, doc='A path to the project that you\'re wanting to run.'),
        'options':       GangaFileItem(defvalue=[], sequence=1, doc='List of files which contain the options I want to pass to gaudirun.py'),
        'uploadedInput': GangaFileItem(defvalue=None, hidden=1, doc='This stores the input for the job which has been pre-uploaded so that it gets to the WN'),
        'jobScriptArchive': GangaFileItem(defvalue=None, hidden=1, copyable=0, doc='This file stores the uploaded scripts which are generated fron this app to run on the WN'),
        'useGaudiRun':  SimpleItem(defvalue=True, doc='Should \'options\' be run as "python options.py data.py" rather than "gaudirun.py options.py data.py"'),
        'platform' :    SimpleItem(defvalue='x86_64-slc6-gcc62-opt', typelist=[str], doc='Platform the application was built for'),
        'autoDBtags' :  SimpleItem(defvalue=False, doc='Automatically set database tags for MC'),
        'extraOpts':    SimpleItem(defvalue='', typelist=[str], doc='An additional string which is to be added to \'options\' when submitting the job'),
        'extraArgs':    SimpleItem(defvalue=[], typelist=[str], sequence=1, doc='Extra runtime arguments which are passed to the code running on the WN'),
        'getMetadata':  SimpleItem(defvalue=False, doc='Do you want to get the metadata from your jobs'),

        # Prepared job object
        'is_prepared':  SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, hidden=0, typelist=[None, ShareDir], protected=0, comparable=1,
            doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        'hash':         SimpleItem(defvalue=None, typelist=[None, str], hidden=1, doc='MD5 hash of the string representation of applications preparable attributes'),
        'envVars':      SimpleItem(defvalue=None, typelist=[None, dict], hidden=1, doc='A dict to store the environment variable "XMLSUMMARYBASEROOT" for use when merging the XML summary'),
        })
    _category = 'applications'
    _name = 'GaudiExec'
    _exportmethods = ['prepare', 'unprepare', 'execCmd', 'readInputData']

    cmake_sandbox_name = 'cmake-input-sandbox.tgz'
    build_target = 'ganga-input-sandbox'
    build_dest = 'input-sandbox.tgz'
    sharedOptsFile_baseName = 'jobScripts-%s.tar'


    def __setattr__(self, attr, value):
        """
        This overloads the baseclass setter method and allows for dynamic evaluation of a parameter on assignment
        Args:
            attr (str): Name of the attribute which is being assigned for this class
            value (unknown): The raw value which is being passed to this class for assigning to the attribute
        """

        actual_value = value
        if attr == 'directory':
            if value:
                actual_value = path.abspath(fullpath(expandfilename(value)))
        elif attr == 'options':
            if isinstance(value, str):
                new_file = allComponentFilters['gangafiles'](value, None)
                actual_value = [ new_file ]
            elif isinstance(value, IGangaFile):
                actual_value = [ value ]
            elif not isinstance(value, (list, tuple, GangaList, type(None))):
                logger.warning("Possibly setting wrong type for options: '%s'" % type(value))

        super(GaudiExec, self).__setattr__(attr, actual_value)


    def unprepare(self, force=False):
        """
        Unprepare the GaudiExec App
        Args:
            force (bool): Forces an un-prepare
        """
        logger.debug('Running unprepare in GaudiExec app')
        if self.is_prepared is not None:
            self.decrementShareCounter(self.is_prepared)
            self.is_prepared = None
        self.hash = None
        self.uploadedInput = None
        self.jobScriptArchive = None

    @gaudiExecBuildLock
    def prepare(self, force=False):
        """
        This method creates a set of prepared files for the application to pass to the RTHandler
        Args:
            force (bool): Forces a prepare to be run
        """

        if (self.is_prepared is not None) and not force:
            raise ApplicationPrepareError('%s application has already been prepared. Use prepare(force=True) to prepare again.' % getName(self))

        # lets use the same criteria as the configure() method for checking file existence & sanity
        # this will bail us out of prepare if there's somthing odd with the job config - like the executable
        # file is unspecified, has a space or is a relative path
        self.configure(self)
        logger.info('Preparing %s application.' % getName(self))

        this_build_target = self.buildGangaTarget()

        self.is_prepared = ShareDir()
        logger.info('Created shared directory: %s' % (self.is_prepared.name))
        try:
            # copy any 'preparable' objects into the shared directory
            send_to_sharedir = self.copyPreparables()
            # add the newly created shared directory into the metadata system
            # if the app is associated with a persisted object
            self.checkPreparedHasParent(self)

            self.copyIntoPrepDir(this_build_target)
            all_opts_files = self.getOptsFiles()
            for opts_file in all_opts_files:
                if isinstance(opts_file, LocalFile):
                    self.copyIntoPrepDir(path.join( opts_file.localDir, path.basename(opts_file.namePattern) ))
                elif isinstance(opts_file, DiracFile):
                    # NB safe to put it here as should have expressly setup a path for this job by now.
                    # We cannot _not_ place this here based upon the backend.
                    # Always have to put it here regardless of if we're on DIRAC or Local so prepared job can be copied.
                    opts_file.localDir=self.getSharedPath()
                    opts_file.get()
                else:
                    raise ApplicationConfigurationError("Opts file type %s not yet supported please contact Ganga devs if you require this support" % getName(opts_file))
            self.post_prepare()

        except Exception as err:
            logger.debug("Err: %s" % str(err))
            self.unprepare()
            raise

        self.cleanGangaTargetArea(this_build_target)

        return 1


    def getExtraOptsFileName(self):
        """
        Returns the name of the opts file which corresponds to the job which owns this app
        This places the script of interest in a subdir to not overly clutter the WN
        """
        return path.join('opts', 'extra_opts_%s_.py' % self.getJobObject().getFQID('.'))


    def getWrapperScriptName(self):
        """
        Returns the name of the wrapper script file which corresponds to the job which owns this app
        This places the script of interest in a subdir to not overly clutter the WN
        """
        return path.join('wrapper', 'job_%s_optsFileWrapper.py' % self.getJobObject().getFQID('.'))


    def constructExtraFiles(self, job):
        """
        This constructs or appends to an uncompressed archive containing all of the opts files which are required to run on the grid
        Args:
            job (Job): The parent job of this application, we don't care if it's unique or not
        """

        master_job = job.master or job

        df = master_job.application.jobScriptArchive

        folder_dir = master_job.getInputWorkspace(create=True).getPath()

        if not df or df.namePattern == '':
            identifier = '%s_%s' % (master_job.id, uuid.uuid4())
            unique_name = GaudiExec.sharedOptsFile_baseName % identifier
            master_job.application.jobScriptArchive = LocalFile(namePattern=unique_name, localDir=folder_dir)
            tar_filename = path.join(folder_dir, unique_name)
            if not path.isfile(tar_filename):
                with tarfile.open(tar_filename, "w"):
                    pass
            with tarfile.open(tar_filename, "a") as tar_file:
                tinfo = tarfile.TarInfo('__timestamp__')
                tinfo.mtime = time.time()
                fileobj = StringIO(getTimestampContent())
                tinfo.size = fileobj.len
                tar_file.addfile(tinfo, fileobj)
        else:
            unique_name = master_job.application.jobScriptArchive.namePattern

        extra_opts_file = self.getExtraOptsFileName()

        # First construct if needed
        if not path.isfile(path.join(folder_dir, unique_name)):
            with tarfile.open(path.join(folder_dir, unique_name), "w"):
                pass

        # Now append the extra_opts file here when needed
        with tarfile.open(path.join(folder_dir, unique_name), "a") as tar_file:
            # Add the extra opts file to the job
            tinfo = tarfile.TarInfo(extra_opts_file)
            tinfo.mtime = time.time()
            fileobj = StringIO(self.extraOpts)
            tinfo.size = fileobj.len
            tar_file.addfile(tinfo, fileobj)

            if not self.useGaudiRun:
                # Add the WN script for wrapping the job
                logger.info("Constructing: %s" % self.getWrapperScriptName())
                tinfo2 = tarfile.TarInfo(self.getWrapperScriptName())
                tinfo2.mtime = time.time()
                fileobj2 = StringIO(self.getWNPythonContents())
                tinfo2.size = fileobj2.len
                tar_file.addfile(tinfo2, fileobj2)


    def cleanGangaTargetArea(self, this_build_target):
        """
        Method to remove the build target and other files not needed to reproduce the same build target again
        Args:
            this_build_target (str): This is the full path of the build target
        """
        logger.debug("Cleaning up area after prepare")

        # Don't delete these
        preserved_set = set(['run'])

        build_dir = path.dirname(this_build_target)
        for obj in set(listdir(build_dir)) - preserved_set:
            logger.debug("del: %s of %s" %(obj, set(listdir(build_dir)) - preserved_set))
            if path.isfile(path.join(build_dir, obj)):
                unlink(path.join(build_dir, obj))
            elif path.isdir(path.join(build_dir, obj)):
                shutil.rmtree(path.join(build_dir, obj), ignore_errors=True)


    def configure(self, masterappconfig):
        """
        Required even though nothing is done in this step for this App
        Args:
            masterappconfig (unknown): This is the output from the master_configure from the parent app
        """
        # Lets test the inputs
        opt_file = self.getOptsFiles()
        dir_name = self.directory
        return (None, None)

    def getOptsFiles(self):
        """
        This function returns a sanitized absolute path to the self.options file from user input
        """
        for this_opt in self.options:
            if isinstance(this_opt, LocalFile):
                ## FIXME LocalFile should return the basename and folder in 2 attibutes so we can piece it together, now it doesn't
                full_path = path.join(this_opt.localDir, this_opt.namePattern)
                if not path.exists(full_path):
                    raise ApplicationConfigurationError("Opts File: \'%s\' has been specified but does not exist please check and try again!" % full_path)
            elif isinstance(this_opt, DiracFile):
                pass
            else:
                logger.error("opts: %s" % self.options)
                raise ApplicationConfigurationError("Opts file type %s not yet supported please contact Ganga devs if you require this support" % getName(this_opt))

        if self.options or self.extraOpts:
            return self.options
        else:
            raise ApplicationConfigurationError("No options (as options files or extra options) has been specified. Please provide some.")

    def getEnvScript(self):
        """
        Return the script which wraps the running command in a correct environment
        """
        return 'export CMTCONFIG=%s; source LbLogin.sh --cmtconfig=%s && ' % (self.platform, self.platform)


    def getWNEnvScript(self):
        """
        Return the script to setup the correct env on a WN
        """
        return 'export CMTCONFIG=%s; source $LHCb_release_area/LBSCRIPTS/prod/InstallArea/scripts/LbLogin.sh --cmtconfig=%s && ' % (self.platform, self.platform)

    def execCmd(self, cmd):
        """
        This method executes a command within the namespace of the project. The cmd is placed in a bash script which is executed within the env
        This will adopt the platform associated with this application.

        e.g. The following will execute a 'make' command within the given project dir

            app = GaudiExec('some/path')
            app.execCmd('make')

        Args:
            cmd (str): This is the command(s) which are to be executed within the project environment and directory
        """
        if not self.directory:
            raise GangaException("Cannot run a command using GaudiExec without a directory first being set!")
        if not path.isdir(self.directory):
            raise GangaException("The given directory: '%s' doesn't exist!" % self.directory)

        cmd_file = tempfile.NamedTemporaryFile(suffix='.sh', delete=False)
        if not cmd.startswith('./run '):
            cmd = './run ' + cmd

        cmd_file.write("#!/bin/bash")
        cmd_file.write("\n")
        cmd_file.write(self.getEnvScript())
        cmd_file.write(cmd)
        cmd_file.flush()
        cmd_file.close()
        st = os_stat(cmd_file.name)
        chmod(cmd_file.name, st.st_mode | stat.S_IEXEC)

        logger.debug("Running: %s" % cmd_file.name)

        # I would have preferred to execute all commands against inside `./run` so we have some sane behaviour
        # but this requires a build to have been run before we can use this command reliably... so we're just going to be explicit

        if not path.isfile(path.join(self.directory, 'build.%s' %self.platform, 'run')):
            initialCommand = 'export CMTCONFIG=%s && source LbLogin.sh --cmtconfig=%s && make' % (self.platform, self.platform)
            rc, stdout, stderr = _exec_cmd(initialCommand, self.directory)
            if rc != 0:
                logger.error("Failed to perform initial make on a Cmake based project")
                logger.error("This is required so that the './run' target exists and is callable within the project")
                logger.error("StdErr: %s" % str(stderr))
                raise GangaException("Failed to execute command")
            if cmd != 'make':
                rc, stdout, stderr = _exec_cmd(cmd_file.name, self.directory)
        else:
            rc, stdout, stderr = _exec_cmd(cmd_file.name, self.directory)

        if rc != 0:
            logger.error("Failed to execute command: %s" % cmd_file.name)
            logger.error("Tried to execute command in: %s" % self.directory)
            logger.error("StdErr: %s" % str(stdout))
            raise GangaException("Failed to Execute command")

        unlink(cmd_file.name)

        return rc, stdout, stderr


    def buildGangaTarget(self):
        """
        This builds the ganga target 'ganga-input-sandbox' for the project defined by self.directory
        This returns the absolute path to the file after it has been created. It will fail if things go wrong or the file fails to generate
        """
        logger.info("Make-ing target '%s'     (This may take a few minutes depending on the size of your project)" % GaudiExec.build_target)
        # Up to the user to run something like make clean... (Although that would avoid some potential CMake problems)
        self.execCmd('make %s' % GaudiExec.build_target)

        targetPath = path.join(self.directory, 'build.%s' % self.platform, 'ganga')
        if not path.isdir(targetPath):
            raise GangaException("Target Path: %s NOT found!" % targetPath)
        sandbox_str = '%s' % GaudiExec.build_dest
        targetFile = path.join(targetPath, sandbox_str)
        if not path.isfile(targetFile):
            raise GangaException("Target File: %s NOT found!" % targetFile)
        wantedTargetFile = path.join(targetPath, GaudiExec.cmake_sandbox_name)
        rename(targetFile, wantedTargetFile)
        if not path.isfile(wantedTargetFile):
            raise GangaException("Wanted Target File: %s NOT found" % wantedTargetFile)

        logger.info("Built %s" % wantedTargetFile)


        # FIXME Make this more pythonic or have a common method for getting the env rather than re-inventing the wheel

        # Whilst we are here let's store the application environment but ignore awkward ones
        env, envstdout, envstderr = self.execCmd('./run env')
        envDict = {}
        for item in envstdout.split("\n"):
            if len(item.split("="))==2:
                if item.split("=")[0] == 'XMLSUMMARYBASEROOT':
                    envDict[item.split("=")[0]] = item.split("=")[1]
        self.envVars = envDict

        return wantedTargetFile


    def postprocess(self):
        from GangaLHCb.Lib.Applications import XMLPostProcessor
        if self.getMetadata:
            XMLPostProcessor.GaudiExecPostProcess(self, logger)

        #Remove one of the replicas for the job script archive and cmake tarball
        if not self.getJobObject().master:
            self.removeUploadedReplicas()

    def removeUploadedReplicas(self):
        """
        Remove all the replicas of the cmake tarball and job script archive. Will leave
        one of each at a random location.
        """
        #Start with the job script archive

        if isinstance(self.jobScriptArchive, DiracFile):
            self.jobScriptArchive.getReplicas()
            while len(self.jobScriptArchive.locations) > 1:
                SEToRemove = random.choice(self.jobScriptArchive.locations)
                self.jobScriptArchive.removeReplica(SEToRemove)

        if isinstance(self.uploadedInput, DiracFile):
            self.uploadedInput.getReplicas()
            while len(self.uploadedInput.locations) > 1:
                SEToRemove = random.choice(self.uploadedInput.locations)
                self.uploadedInput.removeReplica(SEToRemove)

    def getenv(self, cache_env=False):
        """
        A function to return the environment of the built application
        """
        defaultXMLBASE = '/cvmfs/lhcb.cern.ch/lib/lhcb/LHCB/LHCB_v42r4/Kernel/XMLSummaryBase'
        if self.envVars:
            return self.envVars
        else:
            logger.debug('Using default value for XMLSUMMARYBASE: %s', defaultXMLBASE)
            return {'XMLSUMMARYBASEROOT' : defaultXMLBASE} 

    def readInputData(self, opts):
        """
        This reads the inputdata from a file and assigns it to the inputdata field of the parent job.

        Or you can use BKQuery and the box repo to save having to do this over and over
        Args:
            opts (str): This is the file which contains the inputdata we want to read in
        """
        input_dataset = getGaudiExecInputData(opts, self)
        try:
            job = self.getJobObject()
        except AssertionError:
            raise GangaException("This makes no sense without first belonging to a job object as I can't assign input data!")

        if job.inputdata is not None and len(job.inputdata) > 0:
            logger.warning("Warning Job %s already contained inputdata, overwriting" % job.fqid)

        job.inputdata = input_dataset


    def getWNPythonContents(self):
        """
        Return the wrapper script which is used to run GaudiPython type jobs on the WN
        """
        # FIXME should there be a more central definition of 'data.py' string to rename this for all of LHCb if it ever changes for LHCbDirac
        from ..RTHandlers.GaudiExecRTHandlers import GaudiExecDiracRTHandler

        all_names = [this_o.namePattern  for this_o in self.options]
        return gaudiPythonWrapper(repr(self.extraArgs), self.getExtraOptsFileName(),
                                  GaudiExecDiracRTHandler.data_file, all_names)




from GangaCore.Core.exceptions import GangaException


class CRABServerError(GangaException):
    """Exception for errors regarding CRAB commands."""

    def __init__(self, message=''):
        GangaException.__init__(self, message)
        self.message = message

    def __str__(self):
        return "CRABServerError: %s" % (self.message)



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import os
from GangaDirac.Lib.RTHandlers.DiracRTHUtils import dirac_inputdata, dirac_ouputdata, mangle_job_name, diracAPI_script_template, diracAPI_script_settings, API_nullifier, dirac_outputfile_jdl
from GangaDirac.Lib.Files.DiracFile import DiracFile
from GangaDirac.Lib.RTHandlers.RunTimeHandlerUtils import master_sandbox_prepare, sandbox_prepare, script_generator
from GangaCore.GPIDev.Lib.File.LocalFile import LocalFile
from GangaCore.GPIDev.Lib.File.OutputFileManager import getOutputSandboxPatterns, getWNCodeForOutputPostprocessing
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaCore.GPIDev.Lib.File import File, FileBuffer
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.util import unique
from GangaCore.GPIDev.Base.Proxy import isType, stripProxy
logger = getLogger()
config = getConfig('DIRAC')

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class ExeDiracRTHandler(IRuntimeHandler):

    """The runtime handler to run plain executables on the Dirac backend"""

    def master_prepare(self, app, appmasterconfig):
        inputsandbox, outputsandbox = master_sandbox_prepare(app, appmasterconfig)
        if type(app.exe) == File:
            input_dir = app.getJobObject().getInputWorkspace().getPath()
            exefile = os.path.join(input_dir, os.path.basename(app.exe.name))
            if not os.path.exists(exefile):
                msg = 'Executable: "%s" must exist!' % str(exefile)
                raise ApplicationConfigurationError(None, msg)

            os.system('chmod +x %s' % exefile)
        return StandardJobConfig(inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))

    def prepare(self, app, appsubconfig, appmasterconfig, jobmasterconfig):
        inputsandbox, outputsandbox = sandbox_prepare(app, appsubconfig, appmasterconfig, jobmasterconfig)
        input_data,   parametricinput_data = dirac_inputdata(app)
#        outputdata,   outputdata_path      = dirac_ouputdata(app)

        job = stripProxy(app).getJobObject()
        outputfiles = [this_file for this_file in job.outputfiles if isType(this_file, DiracFile)]

        commandline = []
        commandline.append(app.exe)
        if isType(app.exe, File):
            #logger.info("app: %s" % str(app.exe.name))
            #fileName = os.path.join(get_share_path(app), os.path.basename(app.exe.name))
            #logger.info("EXE: %s" % str(fileName))
            #inputsandbox.append(File(name=fileName))
            inputsandbox.append(app.exe)
            commandline[0]=os.path.join('.', os.path.basename(app.exe.name))
        commandline.extend([str(arg) for arg in app.args])
        logger.debug('Command line: %s: ', commandline)

        #exe_script_path = os.path.join(job.getInputWorkspace().getPath(), "exe-script.py")
        exe_script_name = 'exe-script.py'

        logger.info("Setting Command to be: '%s'" % repr(commandline))

        inputsandbox.append(FileBuffer(name=exe_script_name,
                            contents=script_generator(exe_script_template(),
                                                    #remove_unreplaced = False,
                                                    # ,
                                                    COMMAND=repr(commandline),
                                                    OUTPUTFILESINJECTEDCODE = getWNCodeForOutputPostprocessing(job, '    ')
                                                    ),
                                       executable=True))

        contents=script_generator(exe_script_template(), COMMAND=repr(commandline),
                                    OUTPUTFILESINJECTEDCODE = getWNCodeForOutputPostprocessing(job, '    ')
                                    )

        #logger.info("Script is: %s" % str(contents))

        from os.path import abspath, expanduser

        for this_file in job.inputfiles:
            if isinstance(this_file, LocalFile):
                for name in this_file.getFilenameList():
                    inputsandbox.append(File(abspath(expanduser(name))))

        dirac_outputfiles = dirac_outputfile_jdl(outputfiles, config['RequireDefaultSE'])

        # NOTE special case for replicas: replicate string must be empty for no
        # replication
        dirac_script = script_generator(diracAPI_script_template(),
                                        DIRAC_IMPORT='from DIRAC.Interfaces.API.Dirac import Dirac',
                                        DIRAC_JOB_IMPORT='from DIRAC.Interfaces.API.Job import Job',
                                        DIRAC_OBJECT='Dirac()',
                                        JOB_OBJECT='Job()',
                                        NAME=mangle_job_name(app),
                                        # os.path.basename(exe_script_path),
                                        EXE=exe_script_name,
                                        # ' '.join([str(arg) for arg in app.args]),
                                        EXE_ARG_STR='',
                                        EXE_LOG_FILE='Ganga_Executable.log',
                                        ENVIRONMENT=None,  # app.env,
                                        INPUTDATA=input_data,
                                        PARAMETRIC_INPUTDATA=parametricinput_data,
                                        OUTPUT_SANDBOX=API_nullifier(outputsandbox),
                                        OUTPUTFILESSCRIPT=dirac_outputfiles,
                                        OUTPUT_PATH="",  # job.fqid,
                                        SETTINGS=diracAPI_script_settings(app),
                                        DIRAC_OPTS=job.backend.diracOpts,
                                        REPLICATE='True' if config['ReplicateOutputData'] else '',
                                        # leave the sandbox for altering later as needs
                                        # to be done in backend.submit to combine master.
                                        # Note only using 2 #s as auto-remove 3
                                        INPUT_SANDBOX='##INPUT_SANDBOX##'
                                        )

        #logger.info("dirac_script: %s" % dirac_script)

        #logger.info("inbox: %s" % str(unique(inputsandbox)))
        #logger.info("outbox: %s" % str(unique(outputsandbox)))

        return StandardJobConfig(dirac_script,
                                 inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def exe_script_template():
    script_template = """#!/usr/bin/env python
'''Script to run Executable application'''
from __future__ import print_function
from os import system, environ, pathsep, getcwd, listdir, path
import sys
import subprocess

# Main
if __name__ == '__main__':

    my_env = environ.copy()
    my_env['PATH'] = getcwd() + (pathsep + my_env['PATH'])

    exe_cmd = ###COMMAND###

    if isinstance(exe_cmd, str):
        exe_cmd = [exe_cmd]

    if isinstance(exe_cmd, list):
        if path.isfile(path.abspath(exe_cmd[0])):
            exe_cmd[0] = path.abspath(exe_cmd[0])

    err = None
    try:
        rc = subprocess.call(exe_cmd, env=my_env, shell=False)
    except Exception as x:
        rc = -9999
        print('Exception occured in running process: ' + repr(exe_cmd))
        print('Err was: ' + str(x))
        subprocess.call('echo $PATH', shell=True)
        print('PATH: ' + str(my_env['PATH']))
        print('PWD: ' + str(my_env['PWD']))
        print("files on WN: " + str(listdir('.')))
        raise

    print("files on WN: " + str(listdir('.')))

    ###OUTPUTFILESINJECTEDCODE###

    sys.exit(rc)
"""
    return script_template

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
allHandlers.add('Executable', 'Dirac', ExeDiracRTHandler)
